# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/dataset.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&sarus_data_spec/protobuf/dataset.proto\x12\x0fsarus_data_spec\"\x8c\x08\n\x07\x44\x61taset\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12+\n\x04spec\x18\x04 \x01(\x0b\x32\x1d.sarus_data_spec.Dataset.Spec\x12<\n\nproperties\x18\x05 \x03(\x0b\x32(.sarus_data_spec.Dataset.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xca\x02\n\x04Spec\x12;\n\x0btransformed\x18\x01 \x01(\x0b\x32$.sarus_data_spec.Dataset.TransformedH\x00\x12-\n\x04\x66ile\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Dataset.FileH\x00\x12/\n\x05\x66iles\x18\x03 \x01(\x0b\x32\x1e.sarus_data_spec.Dataset.FilesH\x00\x12\x33\n\x07\x61rchive\x18\x04 \x01(\x0b\x32 .sarus_data_spec.Dataset.ArchiveH\x00\x12+\n\x03sql\x18\x05 \x01(\x0b\x32\x1c.sarus_data_spec.Dataset.SqlH\x00\x12;\n\x0bhuggingface\x18\x06 \x01(\x0b\x32$.sarus_data_spec.Dataset.HuggingfaceH\x00\x42\x06\n\x04spec\x1a\xbd\x01\n\x0bTransformed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12Q\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\x38.sarus_data_spec.Dataset.Transformed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a#\n\x04\x46ile\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x0b\n\x03uri\x18\x02 \x01(\t\x1a,\n\x05\x46iles\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x13\n\x0buri_pattern\x18\x02 \x01(\t\x1a>\n\x07\x41rchive\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x16\n\x0e\x61rchive_format\x18\x02 \x01(\t\x12\x0b\n\x03uri\x18\x03 \x01(\t\x1an\n\x03Sql\x12\x0b\n\x03uri\x18\x01 \x01(\t\x12\x32\n\x06tables\x18\x02 \x03(\x0b\x32\".sarus_data_spec.Dataset.Sql.Table\x1a&\n\x05Table\x12\x0e\n\x06schema\x18\x01 \x01(\t\x12\r\n\x05table\x18\x02 \x01(\t\x1a*\n\x0bHuggingface\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05split\x18\x02 \x01(\tb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.dataset_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DATASET_PROPERTIESENTRY._options = None
  _DATASET_PROPERTIESENTRY._serialized_options = b'8\001'
  _DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY._options = None
  _DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_options = b'8\001'
  _DATASET._serialized_start=60
  _DATASET._serialized_end=1096
  _DATASET_PROPERTIESENTRY._serialized_start=219
  _DATASET_PROPERTIESENTRY._serialized_end=268
  _DATASET_SPEC._serialized_start=271
  _DATASET_SPEC._serialized_end=601
  _DATASET_TRANSFORMED._serialized_start=604
  _DATASET_TRANSFORMED._serialized_end=793
  _DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_start=740
  _DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_end=793
  _DATASET_FILE._serialized_start=795
  _DATASET_FILE._serialized_end=830
  _DATASET_FILES._serialized_start=832
  _DATASET_FILES._serialized_end=876
  _DATASET_ARCHIVE._serialized_start=878
  _DATASET_ARCHIVE._serialized_end=940
  _DATASET_SQL._serialized_start=942
  _DATASET_SQL._serialized_end=1052
  _DATASET_SQL_TABLE._serialized_start=1014
  _DATASET_SQL_TABLE._serialized_end=1052
  _DATASET_HUGGINGFACE._serialized_start=1054
  _DATASET_HUGGINGFACE._serialized_end=1096
# @@protoc_insertion_point(module_scope)
