# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/proto_container.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.sarus_data_spec/protobuf/proto_container.proto\x12\x0fsarus_data_spec\x1a\x19google/protobuf/any.proto\"\xae\x01\n\x0eProtoContainer\x12$\n\x06protos\x18\x01 \x03(\x0b\x32\x14.google.protobuf.Any\x12\x43\n\nproperties\x18\x02 \x03(\x0b\x32/.sarus_data_spec.ProtoContainer.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.proto_container_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PROTOCONTAINER_PROPERTIESENTRY._options = None
  _PROTOCONTAINER_PROPERTIESENTRY._serialized_options = b'8\001'
  _PROTOCONTAINER._serialized_start=95
  _PROTOCONTAINER._serialized_end=269
  _PROTOCONTAINER_PROPERTIESENTRY._serialized_start=220
  _PROTOCONTAINER_PROPERTIESENTRY._serialized_end=269
# @@protoc_insertion_point(module_scope)
