syntax = "proto3";
package sarus_data_spec;

message Status {// Current form of the dataset
  string uuid = 1;
  string dataspec = 2;
  string manager = 3;
  string datetime = 4; // ISO 8601 datetime
  map<string, Stage> task_stages = 5; // Various tasks handled by the manager
  map<string, string> properties = 6; // Other properties

  message Stage {
    oneof stage {
      Pending pending = 1;
      Processing processing = 2;
      Ready ready = 3;
      Error error = 4;
    }
    map<string, string> properties = 5;

    message Pending {
    }

    message Processing {
    }

    message Ready {
    }

    message Error {
    }
  }
}
