# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/transform.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import type_pb2 as sarus__data__spec_dot_protobuf_dot_type__pb2
from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2
from sarus_data_spec.protobuf import scalar_pb2 as sarus__data__spec_dot_protobuf_dot_scalar__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(sarus_data_spec/protobuf/transform.proto\x1a#sarus_data_spec/protobuf/type.proto\x1a#sarus_data_spec/protobuf/path.proto\x1a%sarus_data_spec/protobuf/scalar.proto\"\xbb+\n\tTransform\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12\x1d\n\x04spec\x18\x04 \x01(\x0b\x32\x0f.Transform.Spec\x12.\n\nproperties\x18\x05 \x03(\x0b\x32\x1a.Transform.PropertiesEntry\x12\x12\n\ninversible\x18\x06 \x01(\x08\x12\x19\n\x11schema_preserving\x18\x07 \x01(\x08\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xea\x0e\n\x04Spec\x12\'\n\x08identity\x18\x01 \x01(\x0b\x32\x13.Transform.IdentityH\x00\x12\'\n\x08variable\x18\x02 \x01(\x0b\x32\x13.Transform.VariableH\x00\x12\'\n\x08\x63omposed\x18\x03 \x01(\x0b\x32\x13.Transform.ComposedH\x00\x12%\n\x07project\x18\x04 \x01(\x0b\x32\x12.Transform.ProjectH\x00\x12#\n\x06\x66ilter\x18\x05 \x01(\x0b\x32\x11.Transform.FilterH\x00\x12%\n\x07shuffle\x18\x06 \x01(\x0b\x32\x12.Transform.ShuffleH\x00\x12\x1f\n\x04join\x18\x07 \x01(\x0b\x32\x0f.Transform.JoinH\x00\x12\x1f\n\x04\x63\x61st\x18\x08 \x01(\x0b\x32\x0f.Transform.CastH\x00\x12#\n\x06sample\x18\t \x01(\x0b\x32\x11.Transform.SampleH\x00\x12\x30\n\ruser_settings\x18\n \x01(\x0b\x32\x17.Transform.UserSettingsH\x00\x12?\n\x15privacy_unit_tracking\x18\x0b \x01(\x0b\x32\x1e.Transform.PrivacyUnitTrackingH\x00\x12\'\n\x08\x65xternal\x18\x0c \x01(\x0b\x32\x13.Transform.ExternalH\x00\x12)\n\tsynthetic\x18\r \x01(\x0b\x32\x14.Transform.SyntheticH\x00\x12)\n\ttranscode\x18\x0e \x01(\x0b\x32\x14.Transform.TranscodeH\x00\x12\x38\n\x11inverse_transcode\x18\x0f \x01(\x0b\x32\x1b.Transform.InverseTranscodeH\x00\x12&\n\x08get_item\x18\x10 \x01(\x0b\x32\x12.Transform.GetItemH\x00\x12J\n\x1bprivacy_unit_tracking_paths\x18\x11 \x01(\x0b\x32#.Transform.PrivacyUnitTrackingPathsH\x00\x12\x43\n\x17\x61utomatic_user_settings\x18\x12 \x01(\x0b\x32 .Transform.AutomaticUserSettingsH\x00\x12.\n\x0cpublic_paths\x18\x13 \x01(\x0b\x32\x16.Transform.PublicPathsH\x00\x12\x30\n\rassign_budget\x18\x14 \x01(\x0b\x32\x17.Transform.AssignBudgetH\x00\x12\x36\n\x10\x61utomatic_budget\x18\x15 \x01(\x0b\x32\x1a.Transform.AutomaticBudgetH\x00\x12\x37\n\x10\x61ttribute_budget\x18\x16 \x01(\x0b\x32\x1b.Transform.AttributesBudgetH\x00\x12(\n\tsd_budget\x18\x17 \x01(\x0b\x32\x13.Transform.SDBudgetH\x00\x12,\n\x0b\x64\x65rive_seed\x18\x18 \x01(\x0b\x32\x15.Transform.DeriveSeedH\x00\x12+\n\x0bgroup_by_pe\x18\x19 \x01(\x0b\x32\x14.Transform.GroupByPEH\x00\x12\x34\n\x0fsampling_ratios\x18\x1a \x01(\x0b\x32\x19.Transform.SamplingRatiosH\x00\x12*\n\nselect_sql\x18\x1b \x01(\x0b\x32\x14.Transform.SelectSqlH\x00\x12%\n\x07\x65xtract\x18\x1c \x01(\x0b\x32\x12.Transform.ExtractH\x00\x12\x38\n\x11relationship_spec\x18\x1d \x01(\x0b\x32\x1b.Transform.RelationshipSpecH\x00\x12@\n\x15\x64ifferentiated_sample\x18\x1e \x01(\x0b\x32\x1f.Transform.DifferentiatedSampleH\x00\x12;\n\x13validated_user_type\x18\x1f \x01(\x0b\x32\x1c.Transform.ValidatedUserTypeH\x00\x12/\n\rdp_select_sql\x18  \x01(\x0b\x32\x16.Transform.DPSelectSqlH\x00\x12\x36\n\x10\x65rror_estimation\x18! \x01(\x0b\x32\x1a.Transform.ErrorEstimationH\x00\x12(\n\tfit_model\x18\" \x01(\x0b\x32\x13.Transform.FitModelH\x00\x12-\n\x0c\x66it_model_dp\x18# \x01(\x0b\x32\x15.Transform.FitModelDPH\x00\x12;\n\x13generate_from_model\x18$ \x01(\x0b\x32\x1c.Transform.GenerateFromModelH\x00\x12/\n\rto_small_data\x18% \x01(\x0b\x32\x16.Transform.ToSmallDataH\x00\x12&\n\x08push_sql\x18& \x01(\x0b\x32\x12.Transform.PushSQLH\x00\x12.\n\x0cselect_table\x18\' \x01(\x0b\x32\x16.Transform.SelectTableH\x00\x42\x06\n\x04spec\x1a\xf6\x07\n\x08\x45xternal\x12\x11\n\targuments\x18\x01 \x01(\x0c\x12\x17\n\x0fnamed_arguments\x18\x02 \x01(\x0c\x12\x37\n\rop_identifier\x18\x03 \x01(\x0b\x32 .Transform.External.OpIdentifier\x1a\xd7\x04\n\x0cOpIdentifier\x12&\n\x03std\x18\x01 \x01(\x0b\x32\x17.Transform.External.StdH\x00\x12,\n\x06pandas\x18\x02 \x01(\x0b\x32\x1a.Transform.External.PandasH\x00\x12*\n\x05numpy\x18\x03 \x01(\x0b\x32\x19.Transform.External.NumpyH\x00\x12\x34\n\ntensorflow\x18\x04 \x01(\x0b\x32\x1e.Transform.External.TensorflowH\x00\x12.\n\x07sklearn\x18\x05 \x01(\x0b\x32\x1b.Transform.External.SklearnH\x00\x12?\n\x10pandas_profiling\x18\x06 \x01(\x0b\x32#.Transform.External.PandasProfilingH\x00\x12.\n\x07xgboost\x18\x07 \x01(\x0b\x32\x1b.Transform.External.XGBoostH\x00\x12*\n\x05skopt\x18\x08 \x01(\x0b\x32\x19.Transform.External.SkoptH\x00\x12\x30\n\x08imblearn\x18\t \x01(\x0b\x32\x1c.Transform.External.ImblearnH\x00\x12(\n\x04shap\x18\n \x01(\x0b\x32\x18.Transform.External.ShapH\x00\x12*\n\x05scipy\x18\x0b \x01(\x0b\x32\x19.Transform.External.ScipyH\x00\x12\x34\n\noptbinning\x18\x0c \x01(\x0b\x32\x1e.Transform.External.OptBinningH\x00\x42\x04\n\x02op\x1a\x13\n\x03Std\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x16\n\x06Pandas\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Numpy\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x1a\n\nTensorflow\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x17\n\x07Sklearn\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x1f\n\x0fPandasProfiling\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x17\n\x07XGBoost\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Skopt\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x18\n\x08Imblearn\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x14\n\x04Shap\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Scipy\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x1a\n\nOptBinning\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\n\n\x08Identity\x1a*\n\x08Variable\x12\x10\n\x08position\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\x1a\xa9\x01\n\x08\x43omposed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12@\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\'.Transform.Composed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x34\n\x07Project\x12)\n\nprojection\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a/\n\x06\x46ilter\x12%\n\x06\x66ilter\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a\t\n\x07Shuffle\x1a)\n\x04Join\x12!\n\x02on\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a+\n\x04\x43\x61st\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a\x61\n\x06Sample\x12\x12\n\x08\x66raction\x18\x01 \x01(\x01H\x00\x12\x0e\n\x04size\x18\x02 \x01(\x03H\x00\x12%\n\x04seed\x18\x03 \x01(\x0b\x32\x17.sarus_data_spec.ScalarB\x0c\n\nproportion\x1av\n\x0fSchemaInference\x12:\n\x0b\x63\x61st_policy\x18\x01 \x01(\x0e\x32%.Transform.SchemaInference.CastPolicy\"\'\n\nCastPolicy\x12\x08\n\x04NONE\x10\x00\x12\x0f\n\x0bMOST_LIKELY\x10\x01\x1a\x16\n\x07GroupBy\x12\x0b\n\x03key\x18\x01 \x01(\t\x1a[\n\x0bToSmallData\x12\x0c\n\x04size\x18\x01 \x01(\x05\x12\x17\n\x0frandom_sampling\x18\x02 \x01(\x08\x12%\n\x04seed\x18\x03 \x01(\x0b\x32\x17.sarus_data_spec.Scalar\x1a\\\n\x07PushSQL\x12\x1b\n\x13\x64\x61taconnection_name\x18\x01 \x01(\t\x12\x13\n\x0bschema_name\x18\x02 \x01(\t\x12\x12\n\ntable_name\x18\x03 \x01(\t\x12\x0b\n\x03uri\x18\x04 \x01(\t\x1a\x0b\n\tSynthetic\x1a\x0e\n\x0cUserSettings\x1a/\n\x15\x41utomaticUserSettings\x12\x16\n\x0emax_categories\x18\x01 \x01(\x03\x1a\x15\n\x13PrivacyUnitTracking\x1a\x0b\n\tTranscode\x1a\x12\n\x10InverseTranscode\x1ao\n\x14\x44ifferentiatedSample\x12\x12\n\x08\x66raction\x18\x01 \x01(\x01H\x00\x12\x0e\n\x04size\x18\x02 \x01(\x03H\x00\x12%\n\x04seed\x18\x03 \x01(\x0b\x32\x17.sarus_data_spec.ScalarB\x0c\n\nproportion\x1a\x1a\n\x18PrivacyUnitTrackingPaths\x1a\r\n\x0bPublicPaths\x1a.\n\x07GetItem\x12#\n\x04path\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\x1a\x0e\n\x0c\x41ssignBudget\x1a\x11\n\x0f\x41utomaticBudget\x1a\x12\n\x10\x41ttributesBudget\x1a\n\n\x08SDBudget\x1a$\n\nDeriveSeed\x12\x16\n\x0erandom_integer\x18\x01 \x01(\x03\x1a\x0b\n\tGroupByPE\x1a\x10\n\x0eSamplingRatios\x1a\x12\n\x10RelationshipSpec\x1a\x88\x01\n\tSelectSql\x12\x0f\n\x05query\x18\x01 \x01(\tH\x00\x12\x34\n\x0f\x61liased_queries\x18\x02 \x01(\x0b\x32\x19.Transform.AliasedQueriesH\x00\x12*\n\x0bsql_dialect\x18\x03 \x01(\x0e\x32\x15.Transform.SQLDialectB\x08\n\x06select\x1a\x8a\x01\n\x0b\x44PSelectSql\x12\x0f\n\x05query\x18\x01 \x01(\tH\x00\x12\x34\n\x0f\x61liased_queries\x18\x02 \x01(\x0b\x32\x19.Transform.AliasedQueriesH\x00\x12*\n\x0bsql_dialect\x18\x03 \x01(\x0e\x32\x15.Transform.SQLDialectB\x08\n\x06select\x1a@\n\x0e\x41liasedQueries\x12.\n\raliased_query\x18\x01 \x03(\x0b\x32\x17.Transform.AliasedQuery\x1a\x42\n\x0c\x41liasedQuery\x12#\n\x04path\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\x12\r\n\x05query\x18\x02 \x01(\t\x1a\x17\n\x07\x45xtract\x12\x0c\n\x04size\x18\x01 \x01(\x05\x1a\x13\n\x11ValidatedUserType\x1a\x11\n\x0f\x45rrorEstimation\x1aV\n\x08\x46itModel\x12\x12\n\nbatch_size\x18\x01 \x01(\x05\x12\x0e\n\x06\x65pochs\x18\x02 \x01(\x05\x12&\n\ttext_kind\x18\x03 \x01(\x0b\x32\x13.Transform.TextKind\x1aN\n\x08TextKind\x12\x14\n\ntext_field\x18\x01 \x01(\tH\x00\x12\x1f\n\x04\x63hat\x18\x02 \x01(\x0b\x32\x0f.Transform.ChatH\x00\x42\x0b\n\ttext_kind\x1a\x34\n\x04\x43hat\x12\x16\n\x0equestion_field\x18\x01 \x01(\t\x12\x14\n\x0c\x61nswer_field\x18\x02 \x01(\t\x1an\n\nFitModelDP\x12\x12\n\nbatch_size\x18\x01 \x01(\x05\x12\x0e\n\x06\x65pochs\x18\x03 \x01(\x05\x12\x14\n\x0cl2_norm_clip\x18\x04 \x01(\x01\x12&\n\ttext_kind\x18\x05 \x01(\x0b\x32\x13.Transform.TextKind\x1a@\n\x11GenerateFromModel\x12\x16\n\x0emax_new_tokens\x18\x01 \x01(\x05\x12\x13\n\x0btemperature\x18\x02 \x01(\x01\x1a\x32\n\x0bSelectTable\x12#\n\x04path\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\"\x80\x01\n\nSQLDialect\x12\x08\n\x04NONE\x10\x00\x12\x0c\n\x08POSTGRES\x10\x01\x12\x0e\n\nSQL_SERVER\x10\x02\x12\n\n\x06MY_SQL\x10\x03\x12\x0b\n\x07SQLLITE\x10\x04\x12\n\n\x06ORACLE\x10\x05\x12\r\n\tBIG_QUERY\x10\x06\x12\x0c\n\x08REDSHIFT\x10\x07\x12\x08\n\x04HIVE\x10\x08\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.transform_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TRANSFORM_PROPERTIESENTRY._options = None
  _TRANSFORM_PROPERTIESENTRY._serialized_options = b'8\001'
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._options = None
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_options = b'8\001'
  _TRANSFORM._serialized_start=158
  _TRANSFORM._serialized_end=5721
  _TRANSFORM_PROPERTIESENTRY._serialized_start=338
  _TRANSFORM_PROPERTIESENTRY._serialized_end=387
  _TRANSFORM_SPEC._serialized_start=390
  _TRANSFORM_SPEC._serialized_end=2288
  _TRANSFORM_EXTERNAL._serialized_start=2291
  _TRANSFORM_EXTERNAL._serialized_end=3305
  _TRANSFORM_EXTERNAL_OPIDENTIFIER._serialized_start=2405
  _TRANSFORM_EXTERNAL_OPIDENTIFIER._serialized_end=3004
  _TRANSFORM_EXTERNAL_STD._serialized_start=3006
  _TRANSFORM_EXTERNAL_STD._serialized_end=3025
  _TRANSFORM_EXTERNAL_PANDAS._serialized_start=3027
  _TRANSFORM_EXTERNAL_PANDAS._serialized_end=3049
  _TRANSFORM_EXTERNAL_NUMPY._serialized_start=3051
  _TRANSFORM_EXTERNAL_NUMPY._serialized_end=3072
  _TRANSFORM_EXTERNAL_TENSORFLOW._serialized_start=3074
  _TRANSFORM_EXTERNAL_TENSORFLOW._serialized_end=3100
  _TRANSFORM_EXTERNAL_SKLEARN._serialized_start=3102
  _TRANSFORM_EXTERNAL_SKLEARN._serialized_end=3125
  _TRANSFORM_EXTERNAL_PANDASPROFILING._serialized_start=3127
  _TRANSFORM_EXTERNAL_PANDASPROFILING._serialized_end=3158
  _TRANSFORM_EXTERNAL_XGBOOST._serialized_start=3160
  _TRANSFORM_EXTERNAL_XGBOOST._serialized_end=3183
  _TRANSFORM_EXTERNAL_SKOPT._serialized_start=3185
  _TRANSFORM_EXTERNAL_SKOPT._serialized_end=3206
  _TRANSFORM_EXTERNAL_IMBLEARN._serialized_start=3208
  _TRANSFORM_EXTERNAL_IMBLEARN._serialized_end=3232
  _TRANSFORM_EXTERNAL_SHAP._serialized_start=3234
  _TRANSFORM_EXTERNAL_SHAP._serialized_end=3254
  _TRANSFORM_EXTERNAL_SCIPY._serialized_start=3256
  _TRANSFORM_EXTERNAL_SCIPY._serialized_end=3277
  _TRANSFORM_EXTERNAL_OPTBINNING._serialized_start=3279
  _TRANSFORM_EXTERNAL_OPTBINNING._serialized_end=3305
  _TRANSFORM_IDENTITY._serialized_start=3307
  _TRANSFORM_IDENTITY._serialized_end=3317
  _TRANSFORM_VARIABLE._serialized_start=3319
  _TRANSFORM_VARIABLE._serialized_end=3361
  _TRANSFORM_COMPOSED._serialized_start=3364
  _TRANSFORM_COMPOSED._serialized_end=3533
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_start=3480
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_end=3533
  _TRANSFORM_PROJECT._serialized_start=3535
  _TRANSFORM_PROJECT._serialized_end=3587
  _TRANSFORM_FILTER._serialized_start=3589
  _TRANSFORM_FILTER._serialized_end=3636
  _TRANSFORM_SHUFFLE._serialized_start=3638
  _TRANSFORM_SHUFFLE._serialized_end=3647
  _TRANSFORM_JOIN._serialized_start=3649
  _TRANSFORM_JOIN._serialized_end=3690
  _TRANSFORM_CAST._serialized_start=3692
  _TRANSFORM_CAST._serialized_end=3735
  _TRANSFORM_SAMPLE._serialized_start=3737
  _TRANSFORM_SAMPLE._serialized_end=3834
  _TRANSFORM_SCHEMAINFERENCE._serialized_start=3836
  _TRANSFORM_SCHEMAINFERENCE._serialized_end=3954
  _TRANSFORM_SCHEMAINFERENCE_CASTPOLICY._serialized_start=3915
  _TRANSFORM_SCHEMAINFERENCE_CASTPOLICY._serialized_end=3954
  _TRANSFORM_GROUPBY._serialized_start=3956
  _TRANSFORM_GROUPBY._serialized_end=3978
  _TRANSFORM_TOSMALLDATA._serialized_start=3980
  _TRANSFORM_TOSMALLDATA._serialized_end=4071
  _TRANSFORM_PUSHSQL._serialized_start=4073
  _TRANSFORM_PUSHSQL._serialized_end=4165
  _TRANSFORM_SYNTHETIC._serialized_start=4167
  _TRANSFORM_SYNTHETIC._serialized_end=4178
  _TRANSFORM_USERSETTINGS._serialized_start=4180
  _TRANSFORM_USERSETTINGS._serialized_end=4194
  _TRANSFORM_AUTOMATICUSERSETTINGS._serialized_start=4196
  _TRANSFORM_AUTOMATICUSERSETTINGS._serialized_end=4243
  _TRANSFORM_PRIVACYUNITTRACKING._serialized_start=4245
  _TRANSFORM_PRIVACYUNITTRACKING._serialized_end=4266
  _TRANSFORM_TRANSCODE._serialized_start=4268
  _TRANSFORM_TRANSCODE._serialized_end=4279
  _TRANSFORM_INVERSETRANSCODE._serialized_start=4281
  _TRANSFORM_INVERSETRANSCODE._serialized_end=4299
  _TRANSFORM_DIFFERENTIATEDSAMPLE._serialized_start=4301
  _TRANSFORM_DIFFERENTIATEDSAMPLE._serialized_end=4412
  _TRANSFORM_PRIVACYUNITTRACKINGPATHS._serialized_start=4414
  _TRANSFORM_PRIVACYUNITTRACKINGPATHS._serialized_end=4440
  _TRANSFORM_PUBLICPATHS._serialized_start=4442
  _TRANSFORM_PUBLICPATHS._serialized_end=4455
  _TRANSFORM_GETITEM._serialized_start=4457
  _TRANSFORM_GETITEM._serialized_end=4503
  _TRANSFORM_ASSIGNBUDGET._serialized_start=4505
  _TRANSFORM_ASSIGNBUDGET._serialized_end=4519
  _TRANSFORM_AUTOMATICBUDGET._serialized_start=4521
  _TRANSFORM_AUTOMATICBUDGET._serialized_end=4538
  _TRANSFORM_ATTRIBUTESBUDGET._serialized_start=4540
  _TRANSFORM_ATTRIBUTESBUDGET._serialized_end=4558
  _TRANSFORM_SDBUDGET._serialized_start=4560
  _TRANSFORM_SDBUDGET._serialized_end=4570
  _TRANSFORM_DERIVESEED._serialized_start=4572
  _TRANSFORM_DERIVESEED._serialized_end=4608
  _TRANSFORM_GROUPBYPE._serialized_start=4610
  _TRANSFORM_GROUPBYPE._serialized_end=4621
  _TRANSFORM_SAMPLINGRATIOS._serialized_start=4623
  _TRANSFORM_SAMPLINGRATIOS._serialized_end=4639
  _TRANSFORM_RELATIONSHIPSPEC._serialized_start=4641
  _TRANSFORM_RELATIONSHIPSPEC._serialized_end=4659
  _TRANSFORM_SELECTSQL._serialized_start=4662
  _TRANSFORM_SELECTSQL._serialized_end=4798
  _TRANSFORM_DPSELECTSQL._serialized_start=4801
  _TRANSFORM_DPSELECTSQL._serialized_end=4939
  _TRANSFORM_ALIASEDQUERIES._serialized_start=4941
  _TRANSFORM_ALIASEDQUERIES._serialized_end=5005
  _TRANSFORM_ALIASEDQUERY._serialized_start=5007
  _TRANSFORM_ALIASEDQUERY._serialized_end=5073
  _TRANSFORM_EXTRACT._serialized_start=5075
  _TRANSFORM_EXTRACT._serialized_end=5098
  _TRANSFORM_VALIDATEDUSERTYPE._serialized_start=5100
  _TRANSFORM_VALIDATEDUSERTYPE._serialized_end=5119
  _TRANSFORM_ERRORESTIMATION._serialized_start=5121
  _TRANSFORM_ERRORESTIMATION._serialized_end=5138
  _TRANSFORM_FITMODEL._serialized_start=5140
  _TRANSFORM_FITMODEL._serialized_end=5226
  _TRANSFORM_TEXTKIND._serialized_start=5228
  _TRANSFORM_TEXTKIND._serialized_end=5306
  _TRANSFORM_CHAT._serialized_start=5308
  _TRANSFORM_CHAT._serialized_end=5360
  _TRANSFORM_FITMODELDP._serialized_start=5362
  _TRANSFORM_FITMODELDP._serialized_end=5472
  _TRANSFORM_GENERATEFROMMODEL._serialized_start=5474
  _TRANSFORM_GENERATEFROMMODEL._serialized_end=5538
  _TRANSFORM_SELECTTABLE._serialized_start=5540
  _TRANSFORM_SELECTTABLE._serialized_end=5590
  _TRANSFORM_SQLDIALECT._serialized_start=5593
  _TRANSFORM_SQLDIALECT._serialized_end=5721
# @@protoc_insertion_point(module_scope)
