# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/schema.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import type_pb2 as sarus__data__spec_dot_protobuf_dot_type__pb2
from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%sarus_data_spec/protobuf/schema.proto\x12\x0fsarus_data_spec\x1a#sarus_data_spec/protobuf/type.proto\x1a#sarus_data_spec/protobuf/path.proto\"\xd3\x03\n\x06Schema\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0f\n\x07\x64\x61taset\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12#\n\x04type\x18\x04 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12+\n\x0cprivacy_unit\x18\x05 \x01(\x0b\x32\x15.sarus_data_spec.Path\x12;\n\nproperties\x18\x06 \x03(\x0b\x32\'.sarus_data_spec.Schema.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xd9\x01\n\nHypothesis\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0f\n\x07\x64\x61taset\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12#\n\x04type\x18\x04 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12\x46\n\nproperties\x18\x05 \x03(\x0b\x32\x32.sarus_data_spec.Schema.Hypothesis.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.schema_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SCHEMA_PROPERTIESENTRY._options = None
  _SCHEMA_PROPERTIESENTRY._serialized_options = b'8\001'
  _SCHEMA_HYPOTHESIS_PROPERTIESENTRY._options = None
  _SCHEMA_HYPOTHESIS_PROPERTIESENTRY._serialized_options = b'8\001'
  _SCHEMA._serialized_start=133
  _SCHEMA._serialized_end=600
  _SCHEMA_PROPERTIESENTRY._serialized_start=331
  _SCHEMA_PROPERTIESENTRY._serialized_end=380
  _SCHEMA_HYPOTHESIS._serialized_start=383
  _SCHEMA_HYPOTHESIS._serialized_end=600
  _SCHEMA_HYPOTHESIS_PROPERTIESENTRY._serialized_start=331
  _SCHEMA_HYPOTHESIS_PROPERTIESENTRY._serialized_end=380
# @@protoc_insertion_point(module_scope)
