# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/scalar.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%sarus_data_spec/protobuf/scalar.proto\x12\x0fsarus_data_spec\"\xaa\x08\n\x06Scalar\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12*\n\x04spec\x18\x04 \x01(\x0b\x32\x1c.sarus_data_spec.Scalar.Spec\x12;\n\nproperties\x18\x05 \x03(\x0b\x32\'.sarus_data_spec.Scalar.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xbd\x02\n\x04Spec\x12:\n\x0btransformed\x18\x01 \x01(\x0b\x32#.sarus_data_spec.Scalar.TransformedH\x00\x12.\n\x05model\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Scalar.ModelH\x00\x12\x43\n\x0eprivacy_params\x18\x03 \x01(\x0b\x32).sarus_data_spec.Scalar.PrivacyParametersH\x00\x12\x39\n\x0brandom_seed\x18\x04 \x01(\x0b\x32\".sarus_data_spec.Scalar.RandomSeedH\x00\x12\x41\n\x0fsynthetic_model\x18\x05 \x01(\x0b\x32&.sarus_data_spec.Scalar.SyntheticModelH\x00\x42\x06\n\x04spec\x1a\xbc\x01\n\x0bTransformed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12P\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\x37.sarus_data_spec.Scalar.Transformed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xad\x01\n\x05Model\x12I\n\x10pretrained_model\x18\x01 \x01(\x0b\x32-.sarus_data_spec.Scalar.Model.PretrainedModelH\x00\x1aQ\n\x0fPretrainedModel\x12\x1d\n\x15\x66oundation_model_name\x18\x01 \x01(\t\x12\x1f\n\x17peft_initialisation_uri\x18\x02 \x01(\tB\x06\n\x04type\x1a}\n\x11PrivacyParameters\x12?\n\x06points\x18\x01 \x03(\x0b\x32/.sarus_data_spec.Scalar.PrivacyParameters.Point\x1a\'\n\x05Point\x12\x0f\n\x07\x65psilon\x18\x01 \x01(\x01\x12\r\n\x05\x64\x65lta\x18\x02 \x01(\x01\x1a\x1b\n\nRandomSeed\x12\r\n\x05value\x18\x01 \x01(\x05\x1a\x10\n\x0eSyntheticModelb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.scalar_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SCALAR_PROPERTIESENTRY._options = None
  _SCALAR_PROPERTIESENTRY._serialized_options = b'8\001'
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._options = None
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_options = b'8\001'
  _SCALAR._serialized_start=59
  _SCALAR._serialized_end=1125
  _SCALAR_PROPERTIESENTRY._serialized_start=215
  _SCALAR_PROPERTIESENTRY._serialized_end=264
  _SCALAR_SPEC._serialized_start=267
  _SCALAR_SPEC._serialized_end=584
  _SCALAR_TRANSFORMED._serialized_start=587
  _SCALAR_TRANSFORMED._serialized_end=775
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_start=722
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_end=775
  _SCALAR_MODEL._serialized_start=778
  _SCALAR_MODEL._serialized_end=951
  _SCALAR_MODEL_PRETRAINEDMODEL._serialized_start=862
  _SCALAR_MODEL_PRETRAINEDMODEL._serialized_end=943
  _SCALAR_PRIVACYPARAMETERS._serialized_start=953
  _SCALAR_PRIVACYPARAMETERS._serialized_end=1078
  _SCALAR_PRIVACYPARAMETERS_POINT._serialized_start=1039
  _SCALAR_PRIVACYPARAMETERS_POINT._serialized_end=1078
  _SCALAR_RANDOMSEED._serialized_start=1080
  _SCALAR_RANDOMSEED._serialized_end=1107
  _SCALAR_SYNTHETICMODEL._serialized_start=1109
  _SCALAR_SYNTHETICMODEL._serialized_end=1125
# @@protoc_insertion_point(module_scope)
